# Copyright (c) Acconeer AB, 2022
# All rights reserved

ACCONEER_XM_CP2105_MODULE_VID = 0x10C4
ACCONEER_XM_CP2105_MODULE_PID = 0xEA70

STM32L431_ID = 0x0435

DEVICE_NAME_TO_ID = {
    "Accconeer XE125": STM32L431_ID,
    "XE125": STM32L431_ID,
    "XM125": STM32L431_ID,
}
